@@ -17,12 +17,15 @@ use crate::table::Table;
 use crate::thread::Thread;
 use crate::types::{Callback, Integer, LightUserData, LuaRef, Number, RegistryKey};
 use crate::userdata::{AnyUserData, MetaMethod, UserData, UserDataMethods};
+#[cfg(any(rlua_lua53, rlua_lua54))]
+use crate::util::isluainteger;
 use crate::util::{
     assert_stack, callback_error, check_stack, get_userdata, get_wrapped_error,
-    init_userdata_metatable, isluainteger, loadbufferx, pop_error, protect_lua,
-    protect_lua_closure, push_globaltable, push_string, push_userdata_uv, push_wrapped_error,
-    tointegerx, tonumberx, StackGuard,
+    init_userdata_metatable, loadbufferx, pop_error, protect_lua, protect_lua_closure,
+    push_globaltable, push_string, push_userdata_uv, push_wrapped_error, tointegerx, tonumberx,
+    StackGuard,
 };
+
 use crate::value::{FromLua, FromLuaMulti, MultiValue, Nil, ToLua, ToLuaMulti, Value};
 
 #[derive(Copy, Clone, Debug)]
@@ -304,6 +307,7 @@ impl<'lua> Context<'lua> {
     /// Lua manual for details.
     pub fn coerce_integer(self, v: Value<'lua>) -> Result<Option<Integer>> {
         Ok(match v {
+            #[cfg(any(rlua_lua53, rlua_lua54))]
             Value::Integer(i) => Some(i),
             v => unsafe {
                 let _sg = StackGuard::new(self.state);
@@ -472,11 +476,14 @@ impl<'lua> Context<'lua> {
             let _sg = StackGuard::new(self.state);
             assert_stack(self.state, 2);
 
+            #[cfg(any(rlua_lua53, rlua_lua54))]
             ffi::lua_rawgeti(
                 self.state,
                 ffi::LUA_REGISTRYINDEX,
                 key.registry_id as ffi::lua_Integer,
             );
+            #[cfg(any(rlua_lua51))]
+            ffi::lua_rawgeti(self.state, ffi::LUA_REGISTRYINDEX, key.registry_id as c_int);
             self.pop_value()
         };
         T::from_lua(value, self)
@@ -552,6 +559,7 @@ impl<'lua> Context<'lua> {
                 ffi::lua_pushlightuserdata(self.state, ud.0);
             }
 
+            #[cfg(any(rlua_lua53, rlua_lua54))]
             Value::Integer(i) => {
                 ffi::lua_pushinteger(self.state, i);
             }
@@ -609,8 +617,15 @@ impl<'lua> Context<'lua> {
             }
 
             ffi::LUA_TNUMBER => {
+                #[cfg(any(rlua_lua51))]
+                {
+                    let n = Value::Number(ffi::lua_tonumber(self.state, -1));
+                    ffi::lua_pop(self.state, 1);
+                    n
+                }
+                #[cfg(any(rlua_lua53, rlua_lua54))]
                 if isluainteger(self.state, -1) != 0 {
-                    let i = Value::Integer(ffi::lua_tointeger(self.state, -1));
+                    let i = Value::Integer(ffi::lua_tointeger(self.state, -1) as i64);
                     ffi::lua_pop(self.state, 1);
                     i
                 } else {
@@ -797,7 +812,7 @@ impl<'lua> Context<'lua> {
             ffi::lua_setmetatable(self.state, -2);
 
             protect_lua_closure(self.state, 1, 1, |state| {
-                ffi::lua_pushcclosure(state, call_callback, 1);
+                ffi::lua_pushcclosure(state, Some(call_callback), 1);
             })?;
 
             Ok(Function(self.pop_ref()))
@@ -815,12 +830,14 @@ impl<'lua> Context<'lua> {
         let ud_index = self.userdata_metatable::<T>()?;
         let uvalues_count = data.get_uvalues_count();
         push_userdata_uv::<RefCell<T>>(self.state, RefCell::new(data), uvalues_count)?;
-
+        #[cfg(any(rlua_lua53, rlua_lua54))]
         ffi::lua_rawgeti(
             self.state,
             ffi::LUA_REGISTRYINDEX,
             ud_index as ffi::lua_Integer,
         );
+        #[cfg(any(rlua_lua51))]
+        ffi::lua_rawgeti(self.state, ffi::LUA_REGISTRYINDEX, ud_index as c_int);
         ffi::lua_setmetatable(self.state, -2);
 
         Ok(AnyUserData(self.pop_ref()))
@@ -839,10 +856,16 @@ impl<'lua> Context<'lua> {
         source: &[u8],
         name: Option<&CString>,
         env: Option<Value<'lua>>,
+        allow_binary: bool,
     ) -> Result<Function<'lua>> {
         unsafe {
             let _sg = StackGuard::new(self.state);
             assert_stack(self.state, 1);
+            let mode = if allow_binary {
+                cstr!("bt")
+            } else {
+                cstr!("t")
+            };
 
             match if let Some(name) = name {
                 loadbufferx(
@@ -850,7 +873,7 @@ impl<'lua> Context<'lua> {
                     source.as_ptr() as *const c_char,
                     source.len(),
                     name.as_ptr() as *const c_char,
-                    cstr!("t"),
+                    mode,
                 )
             } else {
                 loadbufferx(
@@ -858,7 +881,7 @@ impl<'lua> Context<'lua> {
                     source.as_ptr() as *const c_char,
                     source.len(),
                     ptr::null(),
-                    cstr!("t"),
+                    mode,
                 )
             } {
                 ffi::LUA_OK => {
@@ -939,10 +962,12 @@ impl<'lua, 'a> Chunk<'lua, 'a> {
         // actual lua repl does.
         let mut expression_source = b"return ".to_vec();
         expression_source.extend(self.source);
-        if let Ok(function) =
-            self.context
-                .load_chunk(&expression_source, self.name.as_ref(), self.env.clone())
-        {
+        if let Ok(function) = self.context.load_chunk(
+            &expression_source,
+            self.name.as_ref(),
+            self.env.clone(),
+            false,
+        ) {
             function.call(())
         } else {
             self.call(())
@@ -961,7 +986,20 @@ impl<'lua, 'a> Chunk<'lua, 'a> {
     /// This simply compiles the chunk without actually executing it.  
     pub fn into_function(self) -> Result<Function<'lua>> {
         self.context
-            .load_chunk(self.source, self.name.as_ref(), self.env)
+            .load_chunk(self.source, self.name.as_ref(), self.env, false)
+    }
+
+    /// Load this chunk into a regular `Function`.
+    ///
+    /// This simply compiles the chunk without actually executing it.
+    /// Unlike `into_function`, this method allows loading code previously
+    /// compiled and saved with `Function::dump` or `string.dump()`.
+    /// This method is unsafe because there is no check that the precompiled
+    /// Lua code is valid; if it is not this may cause a crash or other
+    /// undefined behaviour.
+    pub unsafe fn into_function_allow_binary(self) -> Result<Function<'lua>> {
+        self.context
+            .load_chunk(self.source, self.name.as_ref(), self.env, true)
     }
 }
 
