@@ -290,7 +290,12 @@ macro_rules! lua_convert_int {
         impl<'lua> ToLua<'lua> for $x {
             fn to_lua(self, _: Context<'lua>) -> Result<Value<'lua>> {
                 if let Some(i) = cast(self) {
-                    Ok(Value::Integer(i))
+                    #[cfg(rlua_lua51)]
+                    let result = Ok(Value::Number(i));
+                    #[cfg(any(rlua_lua53, rlua_lua54))]
+                    let result = Ok(Value::Integer(i));
+
+                    result
                 } else {
                     cast(self)
                         .ok_or_else(|| Error::ToLuaConversionError {
