@@ -1,3 +1,4 @@
+use std::hash::Hash;
 use std::{slice, str};
 
 use crate::error::{Error, Result};
@@ -107,3 +108,11 @@ where
         self.as_bytes() == other.as_ref()
     }
 }
+
+impl<'lua> Eq for String<'lua> {}
+
+impl<'lua> Hash for String<'lua> {
+    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
+        self.as_bytes().hash(state);
+    }
+}
