@@ -74,8 +74,8 @@ pub unsafe fn protect_lua(
 ) -> Result<()> {
     let stack_start = ffi::lua_gettop(state) - nargs;
 
-    ffi::lua_pushcfunction(state, error_traceback);
-    ffi::lua_pushcfunction(state, f);
+    ffi::lua_pushcfunction(state, Some(error_traceback));
+    ffi::lua_pushcfunction(state, Some(f));
     if nargs > 0 {
         rotate(state, stack_start + 1, 2);
     }
@@ -83,7 +83,7 @@ pub unsafe fn protect_lua(
     let ret = ffi::lua_pcall(state, nargs, ffi::LUA_MULTRET, stack_start + 1);
     ffi::lua_remove(state, stack_start + 1);
 
-    if ret == ffi::LUA_OK {
+    if ret == ffi::LUA_OK as i32 {
         Ok(())
     } else {
         Err(pop_error(state, ret))
@@ -96,7 +96,7 @@ pub unsafe fn protect_lua(
 // given function return type is not the return value count, instead the inner function return
 // values are assumed to match the `nresults` param.  Internally uses 3 extra stack spaces, and does
 // not call checkstack.  Provided function must *not* panic, and since it will generally be
-// lonjmping, should not contain any values that implement Drop.
+// longjmping, should not contain any values that implement Drop.
 pub unsafe fn protect_lua_closure<F, R>(
     state: *mut ffi::lua_State,
     nargs: c_int,
@@ -137,8 +137,8 @@ where
 
     let stack_start = ffi::lua_gettop(state) - nargs;
 
-    ffi::lua_pushcfunction(state, error_traceback);
-    ffi::lua_pushcfunction(state, do_call::<F, R>);
+    ffi::lua_pushcfunction(state, Some(error_traceback));
+    ffi::lua_pushcfunction(state, Some(do_call::<F, R>));
     if nargs > 0 {
         rotate(state, stack_start + 1, 2);
     }
@@ -321,7 +321,7 @@ pub unsafe fn init_userdata_metatable<T>(
 
         ffi::lua_pushvalue(state, -1);
         ffi::lua_gettable(state, ffi::lua_upvalueindex(2));
-        if ffi::lua_isnil(state, -1) == 0 {
+        if ffi::lua_isnil(state, -1) == false {
             ffi::lua_insert(state, -3);
             ffi::lua_pop(state, 2);
             1
@@ -355,7 +355,7 @@ pub unsafe fn init_userdata_metatable<T>(
         } else if index_type == ffi::LUA_TFUNCTION {
             ffi::lua_pushvalue(state, members);
             protect_lua_closure(state, 2, 1, |state| {
-                ffi::lua_pushcclosure(state, meta_index_impl, 2);
+                ffi::lua_pushcclosure(state, Some(meta_index_impl), 2);
             })?;
         } else {
             rlua_panic!("improper __index type {}", index_type);
@@ -367,7 +367,7 @@ pub unsafe fn init_userdata_metatable<T>(
     }
 
     push_string(state, "__gc")?;
-    ffi::lua_pushcfunction(state, userdata_destructor::<T>);
+    ffi::lua_pushcfunction(state, Some(userdata_destructor::<T>));
     protect_lua_closure(state, 3, 1, |state| {
         ffi::lua_rawset(state, -3);
     })?;
@@ -556,13 +556,6 @@ pub unsafe fn tonumberx(
 #[cfg(any(rlua_lua53, rlua_lua54))]
 pub use ffi::lua_isinteger as isluainteger;
 
-#[cfg(rlua_lua51)]
-// Implementation of `lua_isinteger()` for Lua 5.1
-pub unsafe fn isluainteger(_state: *mut ffi::lua_State, _index: c_int) -> c_int {
-    // Lua 5.1 doesn't support integers
-    0
-}
-
 #[cfg(any(rlua_lua53, rlua_lua54))]
 pub use ffi::lua_rotate as rotate;
 
@@ -703,6 +696,21 @@ pub unsafe fn loadbufferx(
     ffi::luaL_loadbuffer(state, buf, size, name)
 }
 
+pub unsafe fn dostring(state: *mut ffi::lua_State, s: &str) -> c_int {
+    let load_result = loadbufferx(
+        state,
+        s.as_ptr() as *const c_char,
+        s.len(),
+        cstr!(""),
+        cstr!("t"),
+    );
+    if load_result == ffi::LUA_OK {
+        ffi::lua_pcall(state, 0, ffi::LUA_MULTRET, 0)
+    } else {
+        load_result
+    }
+}
+
 #[cfg(any(rlua_lua53, rlua_lua54))]
 // Like luaL_requiref but doesn't leave the module on the stack.
 pub unsafe fn requiref(
@@ -778,6 +786,19 @@ pub unsafe fn traceback(
     ffi::lua_pushstring(push_state, msg);
 }
 
+#[cfg(any(rlua_lua53, rlua_lua54))]
+pub use ffi::lua_dump as dump;
+
+#[cfg(rlua_lua51)]
+pub unsafe fn dump(
+    state: *mut ffi::lua_State,
+    writer: ffi::lua_Writer,
+    data: *mut c_void,
+    _strip: c_int,
+) -> c_int {
+    ffi::lua_dump(state, writer, data)
+}
+
 // In the context of a lua callback, this will call the given function and if the given function
 // returns an error, *or if the given function panics*, this will result in a call to lua_error (a
 // longjmp).  The error or panic is wrapped in such a way that when calling pop_error back on
@@ -825,14 +846,24 @@ where
             ptr::write(ud as *mut WrappedError, WrappedError(err));
             get_error_metatable(state);
             ffi::lua_setmetatable(state, -2);
-            ffi::lua_error(state)
+            ffi::lua_error(state);
+            unreachable!()
         }
         Err(p) => {
             ffi::lua_settop(state, 1);
             ptr::write(ud as *mut WrappedPanic, WrappedPanic(Some(p)));
-            get_panic_metatable(state);
-            ffi::lua_setmetatable(state, -2);
-            ffi::lua_error(state)
+
+            if get_panic_metatable(state) {
+                ffi::lua_setmetatable(state, -2);
+                ffi::lua_error(state);
+                unreachable!()
+            } else {
+                // The pcall/xpcall wrappers which allow sending a panic
+                // safeul through Lua have not been enabled.
+                // We can't allow a panic to cross the C/Rust boundary, so the
+                // only choice is to abort.
+                std::process::abort()
+            }
         }
     }
 }
@@ -895,7 +926,9 @@ pub unsafe extern "C" fn safe_pcall(state: *mut ffi::lua_State) -> c_int {
     if top == 0 {
         ffi::lua_pushstring(state, cstr!("not enough arguments to pcall"));
         ffi::lua_error(state);
-    } else if ffi::lua_pcall(state, top - 1, ffi::LUA_MULTRET, 0) != ffi::LUA_OK {
+        assert!(false, "code is unreachable");
+        0
+    } else if ffi::lua_pcall(state, top - 1, ffi::LUA_MULTRET, 0) != ffi::LUA_OK as i32 {
         if is_wrapped_panic(state, -1) {
             ffi::lua_error(state);
         }
@@ -933,7 +966,7 @@ pub unsafe extern "C" fn safe_xpcall(state: *mut ffi::lua_State) -> c_int {
     }
 
     ffi::lua_pushvalue(state, 2);
-    ffi::lua_pushcclosure(state, xpcall_msgh, 1);
+    ffi::lua_pushcclosure(state, Some(xpcall_msgh), 1);
     copy(state, 1, 2);
     ffi::lua_replace(state, 1);
 
@@ -990,7 +1023,7 @@ pub unsafe fn get_wrapped_error(state: *mut ffi::lua_State, index: c_int) -> *co
 }
 
 // Initialize the error, panic, and destructed userdata metatables.
-pub unsafe fn init_error_registry(state: *mut ffi::lua_State) {
+pub unsafe fn init_error_registry(state: *mut ffi::lua_State, wrap_panics: bool) {
     assert_stack(state, 8);
 
     // Create error metatable
@@ -1035,11 +1068,11 @@ pub unsafe fn init_error_registry(state: *mut ffi::lua_State) {
     ffi::lua_newtable(state);
 
     ffi::lua_pushstring(state, cstr!("__gc"));
-    ffi::lua_pushcfunction(state, userdata_destructor::<WrappedError>);
+    ffi::lua_pushcfunction(state, Some(userdata_destructor::<WrappedError>));
     ffi::lua_rawset(state, -3);
 
     ffi::lua_pushstring(state, cstr!("__tostring"));
-    ffi::lua_pushcfunction(state, error_tostring);
+    ffi::lua_pushcfunction(state, Some(error_tostring));
     ffi::lua_rawset(state, -3);
 
     ffi::lua_pushstring(state, cstr!("__metatable"));
@@ -1049,22 +1082,23 @@ pub unsafe fn init_error_registry(state: *mut ffi::lua_State) {
     ffi::lua_rawset(state, ffi::LUA_REGISTRYINDEX);
 
     // Create panic metatable
+    if wrap_panics {
+        ffi::lua_pushlightuserdata(
+            state,
+            &PANIC_METATABLE_REGISTRY_KEY as *const u8 as *mut c_void,
+        );
+        ffi::lua_newtable(state);
 
-    ffi::lua_pushlightuserdata(
-        state,
-        &PANIC_METATABLE_REGISTRY_KEY as *const u8 as *mut c_void,
-    );
-    ffi::lua_newtable(state);
+        ffi::lua_pushstring(state, cstr!("__gc"));
+        ffi::lua_pushcfunction(state, Some(userdata_destructor::<WrappedPanic>));
+        ffi::lua_rawset(state, -3);
 
-    ffi::lua_pushstring(state, cstr!("__gc"));
-    ffi::lua_pushcfunction(state, userdata_destructor::<WrappedPanic>);
-    ffi::lua_rawset(state, -3);
+        ffi::lua_pushstring(state, cstr!("__metatable"));
+        ffi::lua_pushboolean(state, 0);
+        ffi::lua_rawset(state, -3);
 
-    ffi::lua_pushstring(state, cstr!("__metatable"));
-    ffi::lua_pushboolean(state, 0);
-    ffi::lua_rawset(state, -3);
-
-    ffi::lua_rawset(state, ffi::LUA_REGISTRYINDEX);
+        ffi::lua_rawset(state, ffi::LUA_REGISTRYINDEX);
+    }
 
     // Create destructed userdata metatable
 
@@ -1113,7 +1147,7 @@ pub unsafe fn init_error_registry(state: *mut ffi::lua_State) {
         cstr!("__ipairs"),
     ] {
         ffi::lua_pushstring(state, method);
-        ffi::lua_pushcfunction(state, destructed_error);
+        ffi::lua_pushcfunction(state, Some(destructed_error));
         ffi::lua_rawset(state, -3);
     }
 
@@ -1129,7 +1163,7 @@ pub unsafe fn init_error_registry(state: *mut ffi::lua_State) {
 
     ffi::lua_newtable(state);
     ffi::lua_pushstring(state, cstr!("__gc"));
-    ffi::lua_pushcfunction(state, userdata_destructor::<String>);
+    ffi::lua_pushcfunction(state, Some(userdata_destructor::<String>));
     ffi::lua_rawset(state, -3);
     ffi::lua_setmetatable(state, -2);
 
@@ -1183,10 +1217,13 @@ unsafe fn is_wrapped_panic(state: *mut ffi::lua_State, index: c_int) -> bool {
         return false;
     }
 
-    get_panic_metatable(state);
-    let res = ffi::lua_rawequal(state, -1, -2) != 0;
-    ffi::lua_pop(state, 2);
-    res
+    if get_panic_metatable(state) {
+        let res = ffi::lua_rawequal(state, -1, -2) != 0;
+        ffi::lua_pop(state, 2);
+        res
+    } else {
+        false
+    }
 }
 
 unsafe fn get_error_metatable(state: *mut ffi::lua_State) {
@@ -1197,12 +1234,31 @@ unsafe fn get_error_metatable(state: *mut ffi::lua_State) {
     ffi::lua_rawget(state, ffi::LUA_REGISTRYINDEX);
 }
 
-unsafe fn get_panic_metatable(state: *mut ffi::lua_State) {
+/// Get the special panic error metatable from the registry.
+///
+/// This may fail if the Lua state was created without the pcall
+/// wrappers.
+///
+/// Returns true if the metatable was pushed to the stack, or false
+/// otherwise (nothing will have been pushed).
+unsafe fn get_panic_metatable(state: *mut ffi::lua_State) -> bool {
     ffi::lua_pushlightuserdata(
         state,
         &PANIC_METATABLE_REGISTRY_KEY as *const u8 as *mut c_void,
     );
-    ffi::lua_rawget(state, ffi::LUA_REGISTRYINDEX);
+    #[cfg(any(rlua_lua53, rlua_lua54))]
+    let mt_type = ffi::lua_rawget(state, ffi::LUA_REGISTRYINDEX);
+    #[cfg(rlua_lua51)]
+    let mt_type = {
+        ffi::lua_rawget(state, ffi::LUA_REGISTRYINDEX);
+        ffi::lua_type(state, -1)
+    };
+    if mt_type == ffi::LUA_TTABLE {
+        true
+    } else {
+        ffi::lua_pop(state, 1);
+        false
+    }
 }
 
 unsafe fn get_destructed_userdata_metatable(state: *mut ffi::lua_State) {
